import matplotlib.pyplot as plt
import math
import numpy as np
import os
import datetime
import shutil
import csv
import warnings
import InitConcIO as ic

class Sub:
    #物質オブジェクトを生成
    def __init__(self, label, conc):
        self.label = label
        self.conc = []
        self.now = [None, None]
        self.k = [None, None, None, None]
        self.dt = None
        self.num = None
        self.velo = Dics({})
        self.init(conc)
    def c(self):
        if self.num == 0:
            return self.now[1]
        elif self.num == 1:
            return self.now[1] + self.k[0]/2*self.dt
        elif self.num == 2:
            return self.now[1] + self.k[1]/2*self.dt
        elif self.num == 3:
            return self.now[1] + self.k[2]*self.dt
    def init(self, conc):
        self.now = [0, conc]
        self.record()
    def record(self):
        self.conc.append(self.now)
    def show(self,path):
        fig = plt.figure()
        x = [x[0] for x in self.conc]
        y = [x[1] for x in self.conc]
        plt.plot(x,y)
        fig.savefig(path + "/" + str(self) + ".png")
        plt.close()
        #csvに書き込み
        with open(path + "/csv/" + str(self) + ".csv", 'w', newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["t", "C(t)"])
            writer.writerows(self.conc)
    def __str__(self):
        return "Sub " + self.label

class Dics:
    def __init__(self, dic):
        self.d = dic
    def __add__(self, other):
        d = {}
        for i in self.d:
            if i in d:
                d[i] += self.d[i]
            else:
                d[i] = self.d[i]
        for i in other.d:
            if i in d:
                d[i] += other.d[i]
            else:
                d[i] = other.d[i]
        return Dics(d)
    def __sub__(self, other):
        d = {}
        for i in self.d:
            if i in d:
                d[i] += self.d[i]
            else:
                d[i] = self.d[i]
        for i in other.d:
            if i in d:
                d[i] -= other.d[i]
            else:
                d[i] = -other.d[i]
        return Dics(d)
    def __neg__(self):
        d = {}
        for i in self.d:
            d[i] = -self.d[i]
        return Dics(d)
    def __str__(self):
        d = self.d
        s = "Velo "
        for i in d:
            s += str(i) + " " + str(d[i]) + ","
        return s

class Velo(Dics):
    #速度オブジェクトを生成
    def __init__(self,label, f = lambda: None, isfast = False):
        self.label = label
        if isfast:
            self.f = lambda: self.k[self.num]
        else:
            self.f = f
        self.d = {self: 1}
        self.isfast = isfast
        self.k = [None, None, None, None]
        self.now = [0, None]
        self.num = 0
        self.velos = []
    def record(self):
        self.velos.append(self.now)
    def show(self,path):
        fig = plt.figure()
        x = [x[0] for x in self.velos]
        y = [x[1] for x in self.velos]
        plt.plot(x,y)
        fig.savefig(path + "/" + str(self) + ".png")
        plt.close()
        #csv書き込み
        with open(path + "/csv/" + str(self) + ".csv", 'w', newline="") as f:
            writer = csv.writer(f)
            writer.writerow(["t", "v(t)"])
            writer.writerows(self.velos)
    def __str__(self):
        return "Velo " + self.label
    def __repr__(self):
        return "Velo " + self.label

class Cell():
    #細胞オブジェクトを生成
    def __init__(self,label):
        self.label = label
    #細胞オブジェクトに物質配列を登録
    def set_subs(self, subs):
        self.subs = subs
        #各subの各veloをsetに登録
        velo_set = set()
        for sub_i in subs:
            for velo_j in sub_i.velo.d:
                velo_set.add(velo_j)
        self.velo_set = velo_set
        #velo_setの各veloを速いものと遅いものに分類して、それぞれ配列に格納
        velo_arr_fast = []
        velo_arr_slow = []
        for velo_i in velo_set:
            if velo_i.isfast:
                velo_arr_fast.append(velo_i)
            else:
                velo_arr_slow.append(velo_i)
        self.velo_arr_fast = velo_arr_fast
        self.velo_arr_slow = velo_arr_slow
        #代謝フラックス解析用の行列A_fastとA_slowを作成する
        A_fast = np.empty((0,len(self.velo_arr_fast)), float)
        A_slow = np.empty((0,len(self.velo_arr_slow)), float)
        for sub_i in self.subs:
            A_fast_i = np.zeros(len(velo_arr_fast))
            A_slow_i = np.zeros(len(velo_arr_slow))
            for velo_j in sub_i.velo.d:
                for k in range(len(velo_arr_fast)):
                    if velo_j.isfast and velo_arr_fast[k] == velo_j:
                        A_fast_i[k] = sub_i.velo.d[velo_j]
                for k in range(len(velo_arr_slow)):
                    if velo_j.isfast==False and velo_arr_slow[k] == velo_j:
                        A_slow_i[k] = sub_i.velo.d[velo_j]
            A_fast = np.append(A_fast, np.array([A_fast_i]), axis=0)
            A_slow = np.append(A_slow, np.array([A_slow_i]), axis=0)
        self.A_fast = A_fast
        self.A_slow = A_slow
        """
        #A_fastが正しく出力されているかチェック
        for sub_i in self.subs:
            print(sub_i)
        print(A_fast , velo_arr_fast)
        """
    def conc(self):
        l = []
        for i in range(len(self.subs)):
            l.append(self.subs[i].c())
        return l
    #Runge-Kutta計算を1step進める
    def step(self,dt):
        #Cellに登録されているすべてのSubの、反応速度を計算してk0~k3に代入
        for n in range(4):
            #cellに登録されているすべてのSubについて、numをnに設定
            for i in range(len(self.subs)):
                self.subs[i].num = n
            #cellに登録されているすべてのveloについて、numをnに設定
            for velo_i in self.velo_set:
                velo_i.num = n
            #v_arr_slowをself.velo_arr_slowから計算し、v_arr_fast=-A_fast^# * A_slow * v_arr_slowを計算
            v_arr_slow = np.zeros(len(self.velo_arr_slow))
            for i, velo_i in enumerate(self.velo_arr_slow):
                v_arr_slow[i] = velo_i.f()
            v_arr_fast = -1*np.dot(np.linalg.pinv(self.A_fast), np.dot(self.A_slow, v_arr_slow))
            #計算したv_arr_fastをvelo_arr_fastの各オブジェクトのk[n]に格納
            for i, velo_i in enumerate(self.velo_arr_fast):
                velo_i.k[n] = v_arr_fast[i]
            for velo_i in self.velo_arr_slow:
                velo_i.k[n] = velo_i.f()
            #Cellに登録されている各Subについてknを計算
            for i in range(len(self.subs)):
                dcdt = 0
                for velo_j in (self.subs[i]).velo.d:
                    dcdt += ((self.subs)[i].velo.d)[velo_j] * velo_j.f()
                self.subs[i].k[n] = dcdt
        #Cellに登録されているすべてのSubの、k0~k3から、nowを計算
        for i in range(len(self.subs)):
            self.subs[i].now = [self.subs[i].now[0]+self.subs[i].dt ,self.subs[i].now[1]+self.subs[i].dt/6 * (self.subs[i].k[0] + self.subs[i].k[1]*2 + self.subs[i].k[2]*2 + self.subs[i].k[3])]
        #Cellに登録されているすべてのVeloの、k0~k3から、nowを計算
        for velo_i in self.velo_set:
            velo_i.now = [velo_i.now[0]+self.dt, (velo_i.k[0]+velo_i.k[1]*2+velo_i.k[2]*2+velo_i.k[3])/6]
    def record(self):
        #Cellに登録されているすべてのsubのrecord関数を呼び出す
        for i in range(len(self.subs)):
            self.subs[i].record()
        #Cellに登録されているすべてのVeloのrecord関数を呼び出す
        for velo_i in self.velo_set:
            velo_i.record()
    def run(self, t_end, dt):
        self.dt = dt
        #Cellに登録されているすべてのSubにdtを設定
        for i in range(len(self.subs)):
            self.subs[i].dt = dt
        for i in range(math.floor(t_end/dt)+1):
            self.step(dt)
            if i%(math.floor(t_end/dt)//100) == 0:
                self.record()
            if i%50 == 0:
                print("\rRunning "+str(i) + "/" + str((math.floor(t_end/dt))),end='')
        self.show()
    def show(self):
        # now = str(datetime.datetime.now().strftime('%Y-%m-%d %H%M%S'))
        now = ""
        # os.mkdir(now + " " + self.label)
        # os.mkdir(now + " " + self.label + "/csv")
        dirname = now + "" + self.label
        if os.path.isdir(dirname):
            warnings.warn("Output directory already exists and will be updated.")
            shutil.rmtree(dirname)
        os.mkdir(dirname)
        os.mkdir(dirname + "/csv")
        #各Subの濃度変化グラフを描画し、そのディレクトリに保存
        for i in range(len(self.subs)):
            # self.subs[i].show(os.getcwd()+"/" + now + " " + self.label)
            self.subs[i].show(os.getcwd()+"/" + dirname)
        #各Velosの速度変化グラフを描画し、そのディレクトリに保存
        for velo_i in self.velo_set:
            # velo_i.show(os.getcwd() + "/" + now + " " + self.label)
            velo_i.show(os.getcwd() + "/" + dirname)

"""
1.すべてのiについて、物質Siを登録。二つ目の引数に初期濃度(単位は自分で固定)を渡す。
Si = Sub("Si", 1.5)

2.すべてのjについて、速度Vjを登録
Veloの第二変数に、速度を表す関数を渡す
速度を表す関数の中に、物質Siの濃度が変数として入るときは、Si.c()と表記する。

Vj = Velo("Vj", lambda n: Si.c()*2+Sj.c()*1)

3.すべてのiについて、物質Siの濃度変化を登録(複数項の足し算、引き算に対応)
Si.velo = Vj - Vi + Vk

4.cellオブジェクトを生成
cell_i = Cell("cell_i")

4-1.cellオブジェクトに物質を登録(cellオブジェクトのsubsメソッドの変数に物質リストを渡す)
cell_i.set_subs([S1,S2,S3])

4-2.cellオブジェクトをt=0 からtまで、dt刻みでシミュレーションする。
cell_i.run(t, dt)

4-3.cellオブジェクトの結果をグラフにし、ファイルに保存する。
cell_i.show()
"""

import sys

filename_temp = sys.argv[1]
InitConcFileName = sys.argv[2]
EndConcFileName = sys.argv[3]
Glc_ext_temp = float(sys.argv[4])
APAP_ext_temp = float(sys.argv[5])
Position_temp = int(sys.argv[6])
Duraion = float(sys.argv[7])
APAP_ext_change_flg = int(sys.argv[8])

for x in [Glc_ext_temp]:
# for x in [4 + 49*0.05, 8 + 21.7*0.05]: # なぜか１回目はステップ数100000で２回目はステップ数10000000。どのようなルールで決まっている？
# for x in [4, 7, 11]:
# for x in [5, 8]:
# for x in [4 + 49*0.05]: # Low Glc
# for x in [8 + 21.7*0.05]: # High Glc
    for y in [Position_temp]:
    # for y in [0, 1, 2]:
    # for y in [0, 2]:
    # for y in [0]: # periportal 門脈
    # for y in [2]: # pericentral 中心静脈
        #periportal(i=0) middle(i=1) pericentral(i=2)の定数
        i = y
        #APAP代謝経路の記述
        APAP = Sub("APAP", 0)
        APAP_ext = Sub("APAP_ext", 0)
        PAPS = Sub("PAPS",  3.16)
        NAPQI = Sub("NAPQI", 0)
        # GSH = Sub("GSH",  [3.28, 1.64, 0.82][i])
        GSH = Sub("GSH",  [2.0, 1.0, 0.5][i])
        Pro_Add = Sub("Pro_Add", 0)
        UDP_GA = Sub("UDP_GA", 0.013)
        # GGC = Sub("GGC", 0)
        GGC = Sub("GGC", 0.38)
        Cys = Sub("Cys", 0)
        B_M = Sub("Beta_M", 0)

        # ----- Read Init Conc -------------------
        initconc = ic.readInitConc(InitConcFileName)
        Cys = Sub("Cys", initconc[0])
        GGC = Sub("GGC", initconc[1])
        B_M = Sub("Beta_M", initconc[2])
        APAP_ext = Sub("APAP_ext", initconc[3])
        APAP = Sub("APAP", initconc[4])
        PAPS = Sub("PAPS", initconc[5])
        NAPQI = Sub("NAPQI", initconc[6])
        # GSH = Sub("GSH", initconc[7])
        # GSH = Sub("GSH",  [3.28, 1.64, 0.82][i])
        GSH = Sub("GSH",  [2.0, 1.0, 0.5][i])
        Pro_Add = Sub("Pro_Add", initconc[8])
        UDP_GA = Sub("UDP_GA", initconc[9])
        # ----------------------------------------

        D_G=0.0833333333333333
        # b_G=x
        b_G=0.5
        # k_GSH=[139,279.252680319093,418][i]
        k_GSH=[139,279.252680319093,418][i]
        k_G=[0.068, 0.124583333333333, 0.19][i]
        # k_S=[0.059 ,0.0394444410950718, 0.0197][i]
        k_S=[0.8 ,0.6, 0.4][i]
        b_S=0.263600374495952
        d_S=0.0833333333333333
        # k_450=[0.875, 1.3125, 1.75][i]
        k_450=[1.2*0.875, 1.2*1.750, 1.2*2.625][i]
        # if x == 4:
        #     k_450 *= 1.2
        # if x == 7:
        #     k_450 *= 1.1
        # k_N=0.0013125
        k_PSH=4.58333333333333
        vmax_GGCS = [2.0, 1.5, 1.0][i]
        vmax_GSHS = [1.5, 1.0, 0.5][i]
        Keq_GGCS = 2
        K_GGCS_Cys = 0.1
        K_GGCS_GGC = 0.1
        KI_GGCS = 0.1
        Keq_GSHS = 10
        K_GSHS_GGC = 0.1
        K_GSHS_GSH = 0.1
        Vmax_Gluc = 0.8*0.4 * 840
        # if x == 4:
        #     Vmax_Gluc *= 0.8
        # if x == 7:
        #     Vmax_Gluc *= 0.9

        V_Sulp = Velo("V_Sulphation", lambda: k_S*PAPS.c()*APAP.c())
        V_Oxid = Velo("V_Oxidation", lambda: k_450*APAP.c())
        V_GSHB = Velo("V_GSH_Binding", lambda: k_GSH*NAPQI.c()*GSH.c())
        V_DPB = Velo("V_DrugProtein_Binding", lambda: k_PSH*NAPQI.c())
        V_Gluc = Velo("V_Glucuronidation", lambda: Vmax_Gluc * k_G * APAP.c() / (APAP.c() + 4) * UDP_GA.c() / (UDP_GA.c() + 0.01))
        V_PAPS = Velo("V_PAPS", lambda: b_S-d_S*PAPS.c())
        #V_GSH = Velo("V_GSH", lambda: b_G-D_G*GSH.c())
        if APAP_ext_change_flg == 1:
            V_APAP_ext = Velo("V_APAP_ext_generation", lambda: APAP_ext_temp if V_APAP_ext.now[0]<0.5 else 0)
        else:
            V_APAP_ext = Velo("V_APAP_ext_generation", lambda: APAP_ext_temp)
        V_APAP_tocell = Velo("V_APAP_tocell", lambda: 50*APAP_ext.c())
        V_APAP_toblood = Velo("V_APAP_toblood", lambda: 50*APAP.c())
        V_Cys = Velo("V_Cys", lambda: b_G-D_G*Cys.c())
        V_PLP = Velo("V_PLP", lambda: 1.0 * Cys.c())
        V_Pyr = Velo("V_Pyr", lambda: 1.0 * B_M.c())
        # V_GGCS = Velo("V_GGCS", lambda: k_GGCS * Cys.c())
        V_GGCS = Velo("V_GGCS", lambda: vmax_GGCS * ( KI_GGCS / ( KI_GGCS + GSH.c() ) ) * ( Cys.c() - GGC.c()/Keq_GGCS ) / (1 + Cys.c() / K_GGCS_Cys + GGC.c() / K_GGCS_GGC ) )
        # V_GSHS = Velo("V_GSHS", lambda: k_GSHS * GGC.c())
        V_GSHS = Velo("V_GSHS", lambda: vmax_GSHS * ( GGC.c() - GSH.c()/Keq_GSHS ) / (1 + GGC.c() / K_GSHS_GGC + GSH.c() / K_GSHS_GSH ) )

        APAP_ext.velo = V_APAP_ext - V_APAP_tocell + V_APAP_toblood
        APAP.velo = -V_Sulp-V_Gluc-V_Oxid + V_APAP_tocell - V_APAP_toblood
        PAPS.velo = -V_Sulp+V_PAPS
        NAPQI.velo = V_Oxid-V_GSHB-V_DPB
        GSH.velo = -V_GSHB + V_GSHS
        Pro_Add.velo = V_DPB
        Cys.velo = V_Cys - V_GGCS - V_PLP
        B_M.velo = V_PLP - V_Pyr
        GGC.velo = V_GGCS - V_GSHS
        #糖代謝経路の記述


        # パラメータ
        ADP = 0.5
        ADP_mito = 7.5
        AMP = 0.16
        ATP = 3.25
        ATP_mito = 17.5
        CO_2 = 5
        CO_2_mito = 5
        # NAD = 1.13
        # NAD_NADH = 1130
        # NAD_mito = 0.046
        # NAD_mito_NADH_mito = 11.5
        P = 5
        P_mito = 8
        PP = 0.008
        # UDP_UDP_Glc_UTP = 1.2
        # Cit = 0.75
        # H = 1 * 10 ** -4
        # H_mito = 6.0 * 10 ** -5
        # Glc_por = 4
        # Ins = 1.55 * (Glc_por) ** 5.7 / (Glc_por ** 5.7 + 7.7 ** 5.7)
        # Glu = 0.253 * (1 - Glc_por ** 5.65 / (Glc_por ** 5.65 + 4.7 ** 5.65)) + 0.02
        #リン酸化率の定数だったかな
        # gam = 1 / 2 * (1 - Ins ** 1.75 / (Ins ** 1.75 + 0.70 ** 1.75) + Glu ** 3.6 / (Glu ** 3.6 + 0.08 ** 3.6))
        # print("gam = ",gam)

        #各種物質
        Glc = Sub("Glc", 4.92)
        DHAP = Sub("DHAP",  4.19 * 10 ** -2)
        Fru16P_2 = Sub("Fru16P_2",  8.08 * 10 ** -4)
        Fru26P_2 = Sub("Fru26P_2",  5.40 * 10 ** -4)
        Fru6P = Sub("Fru6P",  5.70 * 10 ** -2)
        GAP = Sub("GAP",  1.91 * 10 ** -3)
        GDP = Sub("GDP", 1.07 * 10 ** -1)
        GDP_mito = Sub("GDP_mito",  5.6 * 10 ** -1)
        Glc_ER = Sub("Glc_ER", 4.92)
        Glc1P = Sub("Glc1P",  1.17 * 10 ** -2)
        Glc6P = Sub("Glc6P",  1.90 * 10 ** -1)
        Glc6P_ER = Sub("Glc6P_ER",  1.90 * 10 ** -1)
        Glyc = Sub("Glyc", 1.76 * 10)
        GTP = Sub("GTP",  6.93 * 10 ** -1)
        GTP_mito = Sub("GTP_mito",  2.4 * 10 ** -1)
        Lac = Sub("Lac",  6.79 * 10 ** -1)
        Mal = Sub("Mal",  7.2 * 10 ** -1)
        Mal_mito = Sub("Mal_mito", 1.30359361718804)
        OA = Sub("OA", 2.44 * 10 ** -2)
        OA_mito = Sub("OA_mito",  4.647311941222113*10**-4)
        PEP = Sub("PEP",  4.99 * 10 ** -2)
        PEP_mito = Sub("PEP_mito",  1.11 * 10 ** -4)
        _13P2G = Sub("_13P2G",  1.09 * 10 ** -3)
        _2PG = Sub("_2PG", 2.94 * 10 ** -4)
        _3PG = Sub("_3PG",  3.06 * 10 ** -1)
        Pyr = Sub("Pyr",  8.52 * 10 ** -2)
        Pyr_mito = Sub("Pyr_mito",  1.32 * 10 ** -1)
        UDP = Sub("UDP",  1.15)
        UDP_Glc = Sub("UDP_Glc",0.04)
        UTP = Sub("UTP",  7 * 45 * 10 ** -1)
        Glc_ext = Sub("Glc_ext", x)
        Lac_ext = Sub("Lac_ext", 1)

        # ----- Read Init Conc -------------------
        Glc = Sub("Glc", initconc[10])
        DHAP = Sub("DHAP", initconc[11])
        Fru16P_2 = Sub("Fru16P_2", initconc[12])
        Fru26P_2 = Sub("Fru26P_2", initconc[13])
        Fru6P = Sub("Fru6P", initconc[14])
        GAP = Sub("GAP", initconc[15])
        GDP = Sub("GDP", initconc[16])
        GDP_mito = Sub("GDP_mito", initconc[17])
        Glc_ER = Sub("Glc_ER", initconc[18])
        Glc1P = Sub("Glc1P", initconc[19])
        Glc6P = Sub("Glc6P", initconc[20])
        Glc6P_ER = Sub("Glc6P_ER", initconc[21])
        Glyc = Sub("Glyc", initconc[22])
        GTP = Sub("GTP", initconc[23])
        GTP_mito = Sub("GTP_mito", initconc[24])
        Lac = Sub("Lac", initconc[25])
        Mal = Sub("Mal", initconc[26])
        Mal_mito = Sub("Mal_mito", initconc[27])
        OA = Sub("OA", initconc[28])
        OA_mito = Sub("OA_mito", initconc[29])
        PEP = Sub("PEP", initconc[30])
        PEP_mito = Sub("PEP_mito", initconc[31])
        _13P2G = Sub("_13P2G", initconc[32])
        _2PG = Sub("_2PG", initconc[33])
        _3PG = Sub("_3PG", initconc[34])
        Pyr = Sub("Pyr", initconc[35])
        Pyr_mito = Sub("Pyr_mito", initconc[36])
        UDP = Sub("UDP", initconc[37])
        UDP_Glc = Sub("UDP_Glc",initconc[38])
        UTP = Sub("UTP", initconc[39])
        Glc_ext = Sub("Glc_ext", initconc[40])
        Lac_ext = Sub("Lac_ext", initconc[41])
        Glc_ext = Sub("Glc_ext", x)
        # ----------------------------------------
        
        v_0_GK = [3.18*10**3, 4.77*10**3, 6.36*10**3]
        v_max_GP = [1.93*10**2, 1.4*10**2, 8.69*10]
        # if Glc_ext_temp >= 10:
        #     v_max_GP = [0.025*1.93*10**2, 0.1*1.4*10**2, 0.2*8.69*10]
        v_max_GP[0] /= ( 1 + ( 40 - 1 ) * (Glc_ext_temp-4)**4 / ( (Glc_ext_temp-4)**4 + 3**4 ) )
        v_max_GP[1] /= ( 1 + ( 10 - 1 ) * (Glc_ext_temp-4)**4 / ( (Glc_ext_temp-4)**4 + 3**4 ) )
        v_max_GP[2] /= ( 1 + (  5 - 1 ) * (Glc_ext_temp-4)**4 / ( (Glc_ext_temp-4)**4 + 3**4 ) ) 
        v_max_FBP2 = [2.2*3.58*10**2, 2.2*5.37*10**2, 2.2*7.15*10**2]
        v_max_GS = [1.02*10**2, 6.8*10, 3.45*10]
        # if Glc_ext_temp >= 10:
        #     v_max_GS = [40*1.02*10**2, 10*6.8*10, 5*3.45*10]
        v_max_GS[0] *= ( 1 + ( 40 - 1 ) * (Glc_ext_temp-4)**4 / ( (Glc_ext_temp-4)**4 + 3**4 ) )
        v_max_GS[1] *= ( 1 + ( 10 - 1 ) * (Glc_ext_temp-4)**4 / ( (Glc_ext_temp-4)**4 + 3**4 ) )
        v_max_GS[2] *= ( 1 + (  5 - 1 ) * (Glc_ext_temp-4)**4 / ( (Glc_ext_temp-4)**4 + 3**4 ) ) 
        v_max_PEPCK = [7.75*10**2 ,5.11*10**2, 2.43*10**2]
        v_max_G6P_ER = [8.91*10**2, 6.48*10**2, 4.05*10**2]
        v_max_PFK2 = [6.64*10, 9.97*10, 1.33*10**2]
        v_max_PK = [6.13*10**3, 7.45*10**3, 8.77*10**3]
        v_max_UGP = [7.78*10**2, 1.5*7.78*10**2, 2*7.78*10**2]

        v_ALD = Velo("v_ALD", isfast = True)
        v_EN = Velo("v_EN", isfast = True)
        v_FBP1 = Velo("v_FBP1", isfast = True) 
        v_GAPDH = Velo("v_GAPDH", isfast = True)
        v_GK =  Velo("v_GK", lambda: (v_0_GK[i] * Glc.c() / (Glc.c() + 15.9) * (1 - 0.75 * Fru6P.c() / (Fru6P.c() + 0.005))) * ATP * Glc.c() / (ATP + 0.55) / (Glc.c() + 9))
        v_GlcT_ER = Velo("v_GlcT_ER", isfast = True)
        v_GLUT2 = Velo("v_GLUT2", lambda: 9.09 * 10 * ((Glc_ext.c() - Glc.c()) / (1 + Glc_ext.c() / 17.3 + Glc.c() / 17.3)))

        V_0_phospho = 1 / 1.8 / 2.1
        V_max_phospho_GP = V_0_phospho * (AMP / (AMP + 0.017))
        V_0_native = 1 / 2.5 / 500
        V_max_native_GP = V_0_native * (AMP / (AMP + 0.36))
        v_GP = Velo("v_GP", lambda: v_max_GP[i] * (Glyc.c() / 300) * (V_max_phospho_GP * (Glyc.c() * P - Glc1P.c() / 0.21) / ((1 + Glyc.c() / 1.8) * (1 + P / 2.1) + (1 + Glc1P.c() / 0.7) - 1)))
        v_G6P_ER = Velo("v_G6P_ER", lambda: v_max_G6P_ER[i] * Glc6P_ER.c() / (Glc6P_ER.c() + 1.84))
        v_GPI = Velo("v_GPI", isfast = True)
        v_G1PI = Velo("v_G1PI", isfast = True)
        v_G6PT_ER = Velo("v_G6PI_ER", isfast = True)
        v_GS = Velo("v_GS", lambda: (v_max_GS[i] * (300 - Glyc.c()) / (300 - Glyc.c() + 0.1 * 300)) *  (max([UDP_Glc.c(),0]) / (max([UDP_Glc.c(),0]) + 32 * (1 - Glc6P.c() / (Glc6P.c() + 0.09)) + 0.3)))
        v_LacT = Velo("v_LacT", lambda: 5.83 * 10 ** 2 * (Lac_ext.c() - Lac.c()) / (1 + Lac.c() / 2.42 + Lac_ext.c() / 2.42))
        v_LDH = Velo("v_LDH", isfast = True)
        v_MalT = Velo("v_MalT", isfast = True) 
        v_MDH = Velo("v_MDH", isfast = True)
        v_MDH_mito = Velo("v_MDH_mito", isfast = True)
        v_NDK_GTP = Velo("v_NDK_GTP", isfast = True)
        v_NDK_UTP = Velo("v_NDK_UTP", isfast = True)
        v_NDK_GTP_mito = Velo("v_NDK_GTP_mito", isfast = True)
        v_PC = Velo("v_PC", lambda: 3.59 * 10 ** 2 * (ATP_mito * Pyr_mito.c() * CO_2_mito - OA_mito.c() * ADP_mito * P_mito / 6.55) / ( (ATP_mito + 0.14) * (Pyr_mito.c() + 0.33) * (CO_2_mito + 4.2) ) )
        v_PEPCK = Velo("v_PEPCK",lambda: v_max_PEPCK[i] * (OA.c() * GTP.c() - PEP.c() * GDP.c() * CO_2 / 110) / ((1 + OA.c() / 0.024) * (1 + GTP.c() / 0.021) + (1 + PEP.c() / 0.4) * (1 + GDP.c() / 0.02) * (1 + CO_2 / 1.194) - 1))
        v_PEPCK_mito = Velo("v_PEPCK_mito", isfast = True)
        v_PEPT = Velo("v_PEPT", isfast = True)
        v_PFK1 = Velo("v_PFK1", isfast = True) 
        v_PFK2 = Velo("v_PFK2", lambda: v_max_PFK2[i] * Fru6P.c() / (Fru6P.c() + 0.05) * ATP / (ATP + 0.5) * (1 - 0.85 * PEP.c() / (PEP.c() + 0.25)))
        v_FBP2 = Velo("v_FBP2", lambda: v_max_FBP2[i] * (max([Fru26P_2.c(),0]) / (max([Fru26P_2.c(),0]) + 0.002) / (1 + Fru6P.c() / 0.02)))
        v_PGK = Velo("v_PGK", isfast = True) 
        v_PGM = Velo("v_PGM", isfast = True) 
        v_PK = Velo("V_PK", lambda: v_max_PK[i] * (PEP.c() * ADP / (PEP.c() + (5.8 * (1 + ATP / 0.32) * (1 - Fru16P_2.c() / (Fru16P_2.c() + 0.0095)))) / (ADP + 0.33)))
        v_PyrMalT=Velo("v_PyrMalT", isfast = True)
        v_PyrT=Velo("v_PyrT", isfast = True)
        v_TPI = Velo("v_TPI", isfast = True)
        v_UGP = Velo("v_UGP", lambda: v_max_UGP[i] * (UTP.c() * max([Glc1P.c(),0]) - max([UDP_Glc.c(),0])*PP/0.31) / (( 1+UTP.c()/0.2)*(1+max([Glc1P.c(),0])/0.055)*(1+max([UDP_Glc.c(),0])/0.06)*(1+PP/0.084)-1 ) )
        v_UG6D = Velo("v_UG6D", lambda: 50 * (max([UDP_Glc.c(),0]) - UDP_GA.c()) / (max([UDP_Glc.c(),0]) + 1) / (UDP_GA.c() + 1))
        #v_UG6D = Velo("v_UG6D", isfast = True)
        v_UDP_gen = Velo("v_UDP_gen", lambda: 1 * ( 1 - UDP.c() ) )

        #物質に反応速度式を登録
        DHAP.velo = v_ALD - v_TPI
        Fru16P_2.velo = v_PFK1 - v_FBP1 - v_ALD
        Fru26P_2.velo = v_PFK2 - v_FBP2
        Fru6P.velo = v_GPI - v_PFK1 - v_PFK2 + v_FBP1 + v_FBP2
        GAP.velo = v_ALD + v_TPI - v_GAPDH
        GDP.velo = Velo("v_dummy", lambda: 0) # -v_NDK_GTP + v_PEPCK
        GDP_mito.velo = Velo("v_dummy", lambda: 0) # -v_NDK_GTP_mito + v_PEPCK_mito
        Glc.velo = v_GLUT2 - v_GK + v_GlcT_ER
        Glc_ER.velo = v_G6P_ER - v_GlcT_ER
        Glc1P.velo = v_GP - v_G1PI - v_UGP
        Glc6P.velo = v_GK + v_G6PT_ER - v_GPI + v_G1PI
        Glc6P_ER.velo = -v_G6P_ER - v_G6PT_ER
        Glyc.velo = v_GS - v_GP
        GTP.velo = Velo("v_dummy", lambda: 0) # v_NDK_GTP - v_PEPCK
        GTP_mito.velo = Velo("v_dummy", lambda: 0) # v_NDK_GTP_mito - v_PEPCK_mito
        Lac.velo = v_LacT + v_LDH
        Mal.velo = v_MalT - v_MDH + v_PyrMalT
        Mal_mito.velo = -v_MalT - v_MDH_mito - v_PyrMalT
        OA.velo = v_MDH - v_PEPCK
        OA_mito.velo = v_PC - v_PEPCK_mito + v_MDH_mito
        PEP.velo = v_EN - v_PK + v_PEPCK - v_PEPT
        PEP_mito.velo = v_PEPCK_mito + v_PEPT
        _13P2G.velo = v_GAPDH - v_PGK
        _2PG.velo = v_PGM - v_EN
        _3PG.velo = v_PGK - v_PGM
        Pyr.velo = v_PK - v_LDH - v_PyrT - v_PyrMalT + V_Pyr
        Pyr_mito.velo = v_PyrT - v_PC + v_PyrMalT
        UDP.velo = Velo("v_dummy", lambda: 0) # -v_NDK_UTP + v_GS + v_UDP_gen
        UDP_Glc.velo = v_UGP - v_GS - v_UG6D
        UTP.velo = Velo("v_dummy", lambda: 0) # v_NDK_UTP - v_UGP
        UDP_GA.velo = v_UG6D - V_Gluc

        cell1 = Cell(filename_temp)
        cell1.set_subs([Cys, GGC, B_M, APAP_ext, APAP, PAPS, NAPQI, GSH, Pro_Add, UDP_GA, Glc, DHAP, Fru16P_2, Fru26P_2, Fru6P, GAP, GDP, GDP_mito, Glc_ER, Glc1P, Glc6P, Glc6P_ER, Glyc, GTP, GTP_mito, Lac, Mal, Mal_mito, OA, OA_mito, PEP, PEP_mito, _13P2G, _2PG, _3PG, Pyr, Pyr_mito, UDP, UDP_Glc, UTP, Glc_ext, Lac_ext])
        # cell1.run(5,0.00005)
        # cell1.run(5,0.00025)
        # cell1.run(5,0.0005)
        # cell1.run(100,0.0005) # To reach steady state
        # cell1.run(Duraion,0.0005)
        # cell1.run(Duraion,0.00025)
        # cell1.run(Duraion,0.0002)
        cell1.run(Duraion,0.0001)
        
        # ----- Read End Conc --------------------
        EndConc = np.array( [ Cys.c(), GGC.c(), B_M.c(), APAP_ext.c(), APAP.c(),
            PAPS.c(), NAPQI.c(), GSH.c(), Pro_Add.c(), UDP_GA.c(),
            Glc.c(), DHAP.c(), Fru16P_2.c(), Fru26P_2.c(), Fru6P.c(),
            GAP.c(), GDP.c(), GDP_mito.c(), Glc_ER.c(), Glc1P.c(),
            Glc6P.c(), Glc6P_ER.c(), Glyc.c(), GTP.c(), GTP_mito.c(),
            Lac.c(), Mal.c(), Mal_mito.c(), OA.c(), OA_mito.c(),
            PEP.c(), PEP_mito.c(), _13P2G.c(), _2PG.c(), _3PG.c(),
            Pyr.c(), Pyr_mito.c(), UDP.c(), UDP_Glc.c(), UTP.c(),
            Glc_ext.c(), Lac_ext.c() ] )
        ic.writeEndConc(EndConcFileName,EndConc)
        # ----------------------------------------